/*
    SequencerMIDI.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cstdio>
#include <sys/ioctl.h>
#include <iostream>

#include "Sequencer/MIDI.h"
#include "Sequencer/instrument.h"
#include "SequencerMIDI.h"

using namespace std;

SequencerMIDI::SequencerMIDI (State* state_, int midiDeviceNum,
			      Option* option_):
  SequencerOSS (state_, option_), midiDevice (midiDeviceNum)
{
  if (midiDevice < 0 || getNRMidis () <= midiDevice)
    {
      cerr << "SequencerMIDI::SequencerMIDI:midiDeviceNumber is invalid.\n"
	   << "midiDeviceNumber=" << midiDevice
	   << ":NRMidis=" << getNRMidis ()
	   << ".assuming 0." << endl;
      midiDevice = 0;
    }
  assert (0 <= midiDevice);
  assert (midiDevice < getNRMidis ());
}

SequencerMIDI::~SequencerMIDI ()
{
  //do nothing
}

void SequencerMIDI::allNoteOffSoundOff ()
{
  ///////////////////////////////all note off
  char noteOff[3] = {0x90,0x40,0x00};
  sendAllChannel (noteOff);
  //////////////////////////////all sound off
  char soundOff[3] = {0xB0,0x78,0x00};
  sendAllChannel (soundOff);

  for (int i = 0; i < State::MAX_PARTS; i++)
    {
      for (int j = 0; j < Part::MAX_NOTES; j++)
	{
	  traceNoteOff (cacheState.part[i], j);
	}
    }
  dumpBuf ();
  sync ();
}


void SequencerMIDI::reset () throw (Error)
{
  if (ioctl(seqfd, SNDCTL_SEQ_RESET)==-1)//sequencerν
    {
      perror ("seqfd,SNDCTL_SEQ_RESET");
      Error error ("can not reset the midi device");
      throw (error);
    }
  allNoteOffSoundOff ();
  ////////////////////////////reset all controllers
  char reset[3] = {0xB0,0x79,0x00};
  sendAllChannel (reset);
  dumpBuf ();
  sync ();
  
  traceReset ();
}

void SequencerMIDI::sync () throw (Error)
{
  if (ioctl (seqfd,SNDCTL_SEQ_SYNC) == -1)
    {
      perror ("seqfd,SNDCTL_SEQ_SYNC");
      Error error ("can not sync the midi device");
      throw (error);
    }
}

//////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////handle MIDI Event
//////////////////////////////////////////////////////////////////////////
uchar SequencerMIDI::checkChannel (uchar channel)
{
  //assert (channel<MAX_CHANNEL);
  switch (option->portChangeType)
    {
    case Option::IGNORE:
      return channel;
      break;
    case Option::SERIAL:
      if (channel < MIDI::MAX_CHANNEL)
	{
	  if(cacheState.sequencerStatus.port!=SequencerStatus::PORT_A)
	    portChange (SequencerStatus::PORT_A);
	  return channel;
	}
      else if (MIDI::MAX_CHANNEL <= channel &&
	       channel < MIDI::MAX_CHANNEL*2)
	{
	  if (cacheState.sequencerStatus.port != SequencerStatus::PORT_B)
	    portChange (SequencerStatus::PORT_B);
	  return channel-MIDI::MAX_CHANNEL;
	}
      break;
    case Option::USE_ROLAND_SYS_EX:
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!̤
      break;
    }
  
  return channel;
}

//port==0 port_A
//port==1 port_B
void SequencerMIDI::portChange (int port)
{
  //DEBUG_MESSAGE ("SequencerMIDI::portChange port change meta Event found.ingnoring.sorry not supported yet.");
  switch (option->portChangeType)
    {
    case Option::IGNORE:
      //do nothing
      break;
    case Option::SERIAL:
      //0xF5 pp
      //pp<=1 port A,pp>=2 port B
      midiOut (0xF5);
      midiOut ((uchar) port+1);
      break;
    case Option::USE_ROLAND_SYS_EX:
      //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!̤
      break;
    }
  tracePortChange (port);
}

#define MAX_MIDI_CHANNEL 16

void SequencerMIDI::noteOff (uchar channel,uchar note,uchar velocity)
{
  assert (note < 128);
  assert (velocity < 128);
  
  channel=checkChannel (channel);
  if (channel<MAX_MIDI_CHANNEL)
    {
      midiOut (MIDI::NOTE_OFF|channel);
      midiOut (note);
      midiOut (velocity);
      
      traceNoteOff (channel,note,velocity);
    }
}

void SequencerMIDI::noteOn (uchar channel,uchar note,uchar velocity)
{
  assert (note<128);
  assert (velocity<128);
  
  channel=checkChannel (channel);
  //seek,mutenote on̵뤹롣
  if (isPronouce (channel, velocity))
    {//μͥmuteƤʤ
      //velocity == 0λnote off
      if (channel<MAX_MIDI_CHANNEL)
	{
	  midiOut (MIDI::NOTE_ON|channel);
	  midiOut (note);
	  midiOut (velocity);

	  traceNoteOn (channel,note,velocity);
	}
    }
}

void SequencerMIDI::polyphonicKeyPressure
(uchar channel,uchar note,uchar keyPressure)
{
  assert (note<128);
  assert (keyPressure<128);
  
  channel=checkChannel (channel);
  if (channel<MAX_MIDI_CHANNEL)
    {
      midiOut (MIDI::KEY_PRESSURE|channel);
      midiOut (note);
      midiOut (keyPressure);

      tracePolyphonicKeyPressure (channel,note,keyPressure);
    }
}


void SequencerMIDI::controlChange (uchar channel,uchar control,uchar value)
{
  assert (control<128);
  //assert (value<128);
  if (value>=128)
    {
      DEBUG_MESSAGE ("SeqeuncerMIDI::controlChange value>=128!!!");
      value=127;
    }
  
  
  channel=checkChannel (channel);
  if (channel<MAX_MIDI_CHANNEL)
    {
      midiOut (MIDI::CONTROL_CHANGE|channel);
      midiOut (control);
      midiOut (value);
#if 0
      cerr << "controlChange " << hex
	   << (int)(MIDI::CONTROL_CHANGE|channel) << " "
	   << (int) control << " "
	   << (int) value << dec << endl;
#endif

      traceControlChange (channel,control,value);
    }
}

void SequencerMIDI::programChange (uchar channel,uchar program)
{
  assert (program<128);
  
  channel=checkChannel (channel);
  if (channel<MAX_MIDI_CHANNEL)
    {
      midiOut (MIDI::PROGRAM_CHANGE|channel);
      midiOut (program);
#if 0
      cerr << "programChange " << hex
	   << (int)(MIDI::PROGRAM_CHANGE|channel) << " "
	   << (int) program << dec << endl;
#endif

      traceProgramChange (channel,program);
    }
}

void SequencerMIDI::channelPressure (uchar channel,uchar pressure)
{
  assert (pressure<128);
  
  channel=checkChannel (channel);
  if (channel<MAX_MIDI_CHANNEL)
    {
      midiOut (MIDI::CHANNEL_PRESSURE|channel);
      midiOut (pressure);

      traceChannelPressure (channel,pressure);
    }
}

void SequencerMIDI::pitchBendChange (uchar channel,uchar lsb,uchar msb)
{
  assert (lsb<128);
  assert (msb<128);
  
  channel=checkChannel (channel);
  if (channel<MAX_MIDI_CHANNEL)
    {
      midiOut (MIDI::PITCH_BEND|channel);
      midiOut (lsb);
      midiOut (msb);

      tracePitchBendChange (channel,lsb,msb);
    }
}

void SequencerMIDI::exclusive (const uchar* udata,uint32 size,int channel=-1)
{
  //channel=-1̵ѹ
  //ǥͥˤäexclusiveϤݡѹ롣
  //ߤϼס
  
  midiOut (udata,size);

  traceExclusive (udata,size,channel);
}

#undef MAX_MIDI_CHANNEL
